'From Squeak2.9alpha of 13 June 2000 [latest update: #3193] on 6 May 2001 at 9:58:50 am'!
"Change Set:		Hooker-nk
Date:			13 March 2001
Author:			Ned Konz

This implements a global registry for general-purpose
additions of behavior without having to edit code.
For instance, the method changes in the MenuMorph class
in this change set allow one to intercept calls to the various
menu construction functions and add new menu items without
having to edit menu construction code.

However, this can be used to intercept calls to any method.

Lookup of hooks is by category (typically a Symbol, like #menuAdditions), tag (an arbitrary object, often a Symbol), calling class (exact match, no inheritance), and calling selector.

For instance, if we want to catch calls from A>>a to B>>b or C>>c, we would add the line

	self callHooksInCategory: #myCategory.

to both B>>b and C>>c.

Then we would register a global hook using

Object addHook: [ :cat :tag :mc1 :mc2 | 'actions here' ]
			named: #someName
			inCategory: #myCategory
			fromClass: A
			selector: #a.

When callHooksInCategory: call is made from A>>a to either B>>b or C>>c (or any other method that calls self callHooksInCategory: #myCategory), the block given in the registry call above will get called with:
	cat		#myCategory
	tag		#someName
	mc1		A>>a method context (not a copy!!)
	mc2		B>>b or C>>c method context (not a copy!!)

Because mc1 and mc2 are real block contexts, you can skip further execution by using something like:

	thisContext terminateTo: mc2 sender

in your block (in this case, the rest of B>>b or C>>c would be skipped).

Look at Object class>>hookExample1 for an example of adding
an item to the World Menu.

Look at Object class>>hookExample2 for an example of adding
an item to the World Menu, and skipping another one using the terminateTo: trick.
"!

Object subclass: #MMFilter
	instanceVariableNames: 'conditionalString scorePlayerState '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MusicMonitorStuff'!
Object subclass: #MusicMonitor
	instanceVariableNames: 'aScorePlayerMorph MMFilterCollection defaultScorePlayerState '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MusicMonitorStuff'!

!MusicMonitor commentStamp: 'rmp 4/17/2001 18:03' prior: 0!
" The MusicMonitor Class allows you to assign ScorePlayerStates to conditions in Celeste and Scamper.  The Celeste email client and Scamper web browser must be open before the MusicMonitor is initialized. The current ScorePlayerState is determined by the selected instruments, mute buttons, track volumes, track pans, overall rate, and overall volume.  MusicMonitor checks the current state of Celeste when the current email message changes, specifically when Celeste>>setTOCEntry.  MusicMonitor checks the current state of Scamper when the current web page changes, specifically when Scamper>>displayDocument.
"

Celeste current.
Scamper current.

musicMonitor _ MusicMonitor new.
musicMonitor initialize.

" To assign current ScorePlayerState to the default MusicMonitor state"
musicMonitor defaultScorePlayerState.

" To assign current ScorePlayerState to a condition in Scamper do something like: "
musicMonitor filterCurrentStateAndCondition: 
'(Scamper current currentUrl asString) includesSubString: ''squeak'''.

" To assign the current ScorePlayerMorph state to a condition in Celeste do something like: "
musicMonitor filterCurrentStateAndCondition: 
'Celeste current currentMessage subject includesSubString: ''[squeakers]'''.



!

MusicMonitor class
	instanceVariableNames: ''!
AlignmentMorph subclass: #ScorePlayerMorph
	instanceVariableNames: 'scorePlayer trackInstNames instrumentSelector scrollSlider soundFileName LogOfRate '
	classVariableNames: 'LastMIDIPort '
	poolDictionaries: ''
	category: 'Sound-Scores'!
Object subclass: #ScorePlayerState
	instanceVariableNames: 'FileName InstrumentArray VolumeArray BalanceArray MuteArray LogOfRate Volume '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MusicMonitorStuff'!

!Object methodsFor: 'hooks' stamp: 'nk 3/13/2001 21:31'!
callHooksAtDepth: depth inCategory: cat
	"Call all the hooks registered in category cat from my sender's sender (at a depth of depth)"

	| fromClass fromSelector hooks oldss ss |
	ss _ thisContext sender.
	depth timesRepeat: [ oldss _ ss. ss _ ss sender ].
	fromSelector _ ss selector.
	fromClass _ ss receiver class.
	hooks _ (((Hooks at: cat ifAbsent: [ ^nil ])
		at: fromClass ifAbsent: [ ^nil ])
			at: fromSelector ifAbsent: [ ^nil ]).
	hooks keysAndValuesDo: [ :key :block |
		block valueWithArguments: { cat . key . ss home . oldss home }
	]
! !

!Object methodsFor: 'hooks' stamp: 'nk 3/13/2001 15:49'!
callHooksInCategory: cat
	"Call all the hooks registered in the given category for my caller's caller"

	self callHooksAtDepth: 2 inCategory: cat! !


!Celeste methodsFor: 'open-close' stamp: 'rmp 4/17/2001 16:18'!
close
	"Close the mail database."

	userPassword _ nil.
	dependents _ nil.
	mailDB ifNotNil: [
		mailDB close; release.
		mailDB _ nil].
! !

!Celeste methodsFor: 'table of contents pane' stamp: 'rmp 3/28/2001 13:55'!
selectMessage: id
	"Change the currently selected message."

	currentMsgID _ id.
	self changed: #tocEntry.
	self changed: #messageText.

! !

!Celeste methodsFor: 'table of contents pane' stamp: 'rmp 3/28/2001 14:00'!
setTOCEntry: newTOCentry
	"Change the currently selected message. This is done by finding the message ID corresponding to the selected table of contents entry."
	| i |

	(newTOCentry isNil | currentTOC isNil)
		ifTrue: [currentMsgID _ nil]
		ifFalse:
			[i _ currentTOC indexOf: newTOCentry ifAbsent: [nil].
			 (i isNil)
				ifTrue: [currentMsgID _ nil]
				ifFalse: [currentMsgID _ currentMessages at: i]].

	self changed: #tocEntry.
	self changed: #messageText.

	( currentMsgID isNil )
	ifFalse: [ 	
		self trigger: #celesteChangingMessage.
	]! !

!Celeste methodsFor: 'message text pane' stamp: 'rmp 3/28/2001 13:52'!
messageText
	"Answer the text which makes up the complete message (header+body)"
	
	(currentMsgID isNil) ifTrue: [^''].

	"Always show the full message header for messages in the category .tosend. so that all special header lines are preserved, shown and can be edited."
	(currentCategory = '.tosend.')
		ifTrue: [^ mailDB getText: currentMsgID].

	SuppressWorthlessHeaderFields
		ifTrue: [^ self currentMessage formattedText]
		ifFalse: [^ mailDB getText: currentMsgID].
! !

!Celeste methodsFor: 'message text pane' stamp: 'rmp 2/21/2001 12:19'!
messageText: aStringOrText

	currentCategory isNil | currentMsgID isNil ifTrue: [^ self].
	mailDB newText: aStringOrText asString squeakToIso for: currentMsgID.
	self updateTOC.  "in case the message header was changed"
	messageTextView hasUnacceptedEdits: false.
	self changed: #messageText.
	^ true
! !


!MMFilter methodsFor: 'as yet unclassified' stamp: 'rmp 2/28/2001 11:34'!
conditional

	^ conditionalString.! !

!MMFilter methodsFor: 'as yet unclassified' stamp: 'rmp 2/23/2001 14:14'!
conditional: cond

	conditionalString _ cond.! !

!MMFilter methodsFor: 'as yet unclassified' stamp: 'rmp 3/13/2001 16:33'!
scorePlayerState

	^ scorePlayerState.! !

!MMFilter methodsFor: 'as yet unclassified' stamp: 'rmp 3/13/2001 16:33'!
scorePlayerState: state

	scorePlayerState _ state.! !


!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:24'!
add: aString action: aSymbol
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the default target object."
	"Details: Note that the menu item added captures the default target object at the time the item is added; the default target can later be changed before added additional items without affecting the targets of previously added entries. The model is that each entry is like a button that knows everything it needs to perform its action."

	self callHooksInCategory: #menuAdditions.

	self pvtadd: aString
		target: defaultTarget
		selector: aSymbol
		argumentList: EmptyArray.
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:24'!
add: aString selector: aSymbol argument: arg

	self callHooksInCategory: #menuAdditions.

	self pvtadd: aString
		target: defaultTarget
		selector: aSymbol
		argumentList: (Array with: arg)
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:20'!
add: aString subMenu: aMenuMorph
	"Append the given submenu with the given label."

	| item |

	self callHooksInCategory: #menuAdditions.

	item _ MenuItemMorph new.
	item contents: aString;
		subMenu: aMenuMorph.
	self addMorphBack: item.
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:24'!
add: aString target: aTarget action: aSymbol

	self callHooksInCategory: #menuAdditions.

	self pvtadd: aString
		target: aTarget
		selector: aSymbol
		argumentList: EmptyArray
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:24'!
add: aString target: anObject selector: aSymbol
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object."


	self callHooksInCategory: #menuAdditions.

	self pvtadd: aString
		target: anObject
		selector: aSymbol
		argumentList: EmptyArray.
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:24'!
add: aString target: target selector: aSymbol argument: arg
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given argument."


	self callHooksInCategory: #menuAdditions.

	self pvtadd: aString
		target: target
		selector: aSymbol
		argumentList: (Array with: arg)
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:23'!
add: aString target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  Answer the appended menu item."


	self callHooksInCategory: #menuAdditions.
	^self pvtadd: aString target: target selector: aSymbol argumentList: argList
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:21'!
addLine
	"Append a divider line to this menu. Suppress duplicate lines."


	self callHooksInCategory: #menuAdditions.

	submorphs isEmpty ifTrue: [^ self].
	(self lastSubmorph isKindOf: MenuLineMorph)
		ifFalse: [self addMorphBack: MenuLineMorph new].
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:21'!
addList: aList
	"Add the given items to this menu, where each item is a pair (<string> <actionSelector>)..  If an element of the list is simply the symobl $-, add a line to the receiver.  The optional third element of each entry, if present, provides balloon help."


	self callHooksInCategory: #menuAdditions.

	aList do: [:tuple |
		(tuple == #-)
			ifTrue: [self addLine]
			ifFalse:
				[self add: tuple first action: tuple second.
				tuple size > 2 ifTrue:
					[self balloonTextForLastItem: tuple third]]]! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:25'!
addStayUpItem
	"Append a menu item that can be used to toggle this menu's persistent."

	self callHooksInCategory: #menuAdditions.

	self pvtadd: 'keep this menu up'
		target: self
		selector: #toggleStayUp:
		argumentList: EmptyArray.
	self addLine! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:25'!
addStayUpItemSpecial
	"Append a menu item that can be used to toggle this menu's persistent."

	"This variant is resistant to the MVC compatibility in #setInvokingView:"

	self callHooksInCategory: #menuAdditions.

	self pvtadd: 'keep this menu up'
		target: self
		selector: #toggleStayUpIgnore:evt:
		argumentList: #(1).
	self addLine! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:26'!
addUpdating: aWordingSelector action: aSymbol


	self callHooksInCategory: #menuAdditions.

	self pvtaddUpdating: aWordingSelector target: defaultTarget selector: aSymbol argumentList: EmptyArray
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:28'!
addUpdating: aWordingSelector enablement: anEnablementSelector action: aSymbol


	self callHooksInCategory: #menuAdditions.

	self pvtaddUpdating: aWordingSelector enablementSelector: anEnablementSelector target: defaultTarget selector: aSymbol argumentList: EmptyArray
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:28'!
addUpdating: wordingSelector enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is obtained by sending the wordingSelector to the target, and the optional enablementSelector determines whether or not the item should be enabled.  Answer the item itself."

	self callHooksInCategory: #menuAdditions.

	^self pvtaddUpdating: wordingSelector enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:26'!
addUpdating: aWordingSelector target: aTarget action: aSymbol


	self callHooksInCategory: #menuAdditions.

	self pvtaddUpdating: aWordingSelector target: aTarget selector: aSymbol argumentList: EmptyArray
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:30'!
addUpdating: wordingSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is obtained by sending the wordingSelector to the target,  Answer the item added."

	self callHooksInCategory: #menuAdditions.

	^self pvtaddUpdating: wordingSelector target: target selector: aSymbol argumentList: argList
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:32'!
addWithLabel: aLabel enablement: anEnablementSelector action: aSymbol

	self callHooksInCategory: #menuAdditions.

	self pvtaddWithLabel: aLabel enablementSelector: anEnablementSelector target: defaultTarget selector: aSymbol argumentList: EmptyArray
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 3/13/2001 19:31'!
addWithLabel: aLabel enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is constant, and the optional enablementSelector determines whether or not the item should be enabled."

	self callHooksInCategory: #menuAdditions.

	^self pvtaddWithLabel: aLabel enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList! !

!MenuMorph methodsFor: 'private' stamp: 'nk 3/13/2001 16:50'!
pvtadd: aString target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  Answer the appended menu item."

	| item |
	item _ MenuItemMorph new
		contents: aString;
		target: target;
		selector: aSymbol;
		arguments: argList asArray.
	self addMorphBack: item.
	^ item
! !

!MenuMorph methodsFor: 'private' stamp: 'nk 3/13/2001 19:22'!
pvtaddUpdating: wordingSelector enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is obtained by sending the wordingSelector to the target, and the optional enablementSelector determines whether or not the item should be enabled.  Answer the item itself."

	| item |
	item _ UpdatingMenuItemMorph new
		target: target;
		selector: aSymbol;
		wordingProvider: target wordingSelector: wordingSelector;
		enablementSelector: enablementSelector;
		arguments: argList asArray.
	self addMorphBack: item.
	^ item
! !

!MenuMorph methodsFor: 'private' stamp: 'nk 3/13/2001 19:26'!
pvtaddUpdating: wordingSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is obtained by sending the wordingSelector to the target,  Answer the item added."

	| item |
	item _ UpdatingMenuItemMorph new
		target: target;
		selector: aSymbol;
		wordingProvider: target wordingSelector: wordingSelector;
		arguments: argList asArray.
	self addMorphBack: item.
	^ item
! !

!MenuMorph methodsFor: 'private' stamp: 'nk 3/13/2001 19:31'!
pvtaddWithLabel: aLabel enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is constant, and the optional enablementSelector determines whether or not the item should be enabled."

	| item |
	item _ UpdatingMenuItemMorph new
		target: target;
		selector: aSymbol;
		contents: aLabel;
		wordingProvider: target wordingSelector: nil;
		enablementSelector: enablementSelector;
		arguments: argList asArray.
	self addMorphBack: item.
! !


!MessageSend methodsFor: 'evaluating' stamp: 'rmp 3/21/2001 14:40'!
value
	^(arguments isNil)
		ifTrue: [receiver perform: selector]
		ifFalse: [receiver perform: selector withArguments: arguments]! !


!MethodContext methodsFor: 'accessing' stamp: 'nk 3/13/2001 20:25'!
arguments
	"Answer a collection of my arguments"
	^((1 to: self size) collect: [ :i | self at: i ])! !


!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 3/20/2001 20:44'!
MMFilterCollection

	^ MMFilterCollection.! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 3/20/2001 20:46'!
MMFilterCollection: aMMFC

	MMFilterCollection _ aMMFC.! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 4/17/2001 18:18'!
checkMessage: application
	"Check to see if the conditional of any of the filters is true. If so, change to the appropriate state. Otherwise change to the default state"

	| condition changedState |

	changedState _ false.
	1 to: (MMFilterCollection size) do: [:i |
		" check that the current MMFilter has a conditional pertaining to the application
		  that called checkMessage 
		"
		( (MMFilterCollection at: i) conditional includesSubString: application )
		ifTrue: [
			MusicMonitor compile:
				(MMFilterCollection at: i) conditional
					  classified: 'as yet unclassified'.

			condition _ self currentConditional.
			( condition )
			ifTrue: [
				(MMFilterCollection at: i)  scorePlayerState setStateOf: aScorePlayerMorph.
				changedState _ true.
			].
		].
	].

	" If none of the conditionals hold, change to the default state. "
	(changedState)
	ifFalse: [
		(defaultScorePlayerState notNil) ifTrue: [
			defaultScorePlayerState setStateOf: aScorePlayerMorph.
		].
	].! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 4/18/2001 14:40'!
currentConditional ^ Celeste current currentMessage subject includesSubString: '[squeakers]'! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 4/17/2001 18:19'!
defaultScorePlayerState
	" Create a new defaultScorePlayerState and set it to the state of aScorePlayerMorph. "
	defaultScorePlayerState _ ScorePlayerState new.
	defaultScorePlayerState getState: aScorePlayerMorph.
! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 4/18/2001 14:35'!
defaultScorePlayerState: sps

	defaultScorePlayerState _ sps.

! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 4/17/2001 18:21'!
filterCurrentStateAndCondition: aCondition
	" Assign the current ScorePlayerState to aCondition.  If there exists a MMFilter with the same condition,
	  simply change its state.  Otherwise create a new MMFilter with aCondition and the current state and 
	  add it to the MMFilterCollection.
	"
	| mmf sps tempStr |

	tempStr _ 'currentConditional ^ '.
	mmf _ MMFilterCollection detect: [ :c | (c conditional) = (tempStr , aCondition) ] ifNone: [
		mmf _ MMFilter new.	


		mmf conditional: tempStr , aCondition.
		
		MMFilterCollection add: mmf.	
	].

	sps _ ScorePlayerState new.
	sps getState: aScorePlayerMorph.
	mmf scorePlayerState: sps.

! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 4/17/2001 18:23'!
initialize
	"Make sure a ScorePlayer exists and has an open midi file."
	( aScorePlayerMorph = nil )
		ifTrue: [ 
			aScorePlayerMorph _ ScorePlayerMorph new.
			aScorePlayerMorph openInWorld.
		].
	( aScorePlayerMorph scorePlayer trackCount = 0 )
		ifTrue: [
			aScorePlayerMorph openMIDIFile.
		].

	MMFilterCollection _ OrderedCollection new.
	defaultScorePlayerState _ nil.

	" The tells Celeste and Scamper what to send MusicMonitor when they trigger their repsective signals. "
	( Celeste current ) when: #celesteChangingMessage send: #checkMessage: to: self with: 'Celeste'.
	( Scamper current ) when: #scamperChangePage send: #checkMessage: to: self with: 'Scamper'.	! !

!MusicMonitor methodsFor: 'as yet unclassified' stamp: 'rmp 3/13/2001 17:21'!
scorePlayerMorph

	^ aScorePlayerMorph.! !


!Object class methodsFor: 'hooks-documentation' stamp: 'nk 3/13/2001 21:44'!
hookDocumentation
	^self error: 'documentation only'.
"
This implements a global registry for general-purpose
additions of behavior without having to edit code.
For instance, the method changes in the MenuMorph class
in this change set allow one to intercept calls to the various
menu construction functions and add new menu items without
having to edit menu construction code.

However, this can be used to intercept calls to any method.

Lookup of hooks is by category (typically a Symbol, like #menuAdditions), tag (an arbitrary object, often a Symbol), calling class (exact match, no inheritance), and calling selector.

For instance, if we want to catch calls from A>>a to B>>b or C>>c, we would add the line

	self callHooksInCategory: #myCategory.

to both B>>b and C>>c.

Then we would register a global hook using

Object addHook: [ :cat :tag :mc1 :mc2 | 'actions here' ]
			named: #someName
			inCategory: #myCategory
			fromClass: A
			selector: #a.

When callHooksInCategory: call is made from A>>a to either B>>b or C>>c (or any other method that calls self callHooksInCategory: #myCategory), the block given in the registry call above will get called with:
	cat		#myCategory
	tag		#someName
	mc1		A>>a method context (not a copy!!)
	mc2		B>>b or C>>c method context (not a copy!!)

Because mc1 and mc2 are real block contexts, you can skip further execution by using something like:

	thisContext terminateTo: mc2 sender

in your block (in this case, the rest of B>>b or C>>c would be skipped).

Look at Object class>>hookExample1 for an example of adding
an item to the World Menu.

Look at Object class>>hookExample2 for an example of adding
an item to the World Menu, and skipping another one using the terminateTo: trick." ! !

!Object class methodsFor: 'hooks-examples' stamp: 'nk 3/13/2001 21:14'!
hookExample1
	"This adds an item to the World Menu that then removes itself when clicked"
	"Object hookExample1"
	"Hooks explore"

	"This block will be called with:
	cat		#menuAdditions
	tag		#hookExample1
	mc1		TheWorldMenu>>fillIn:from: context
	mc2		menu add* method contexts"

	self addHook: [ :cat :tag :mc1 :mc2 | | str |
				str _ (mc2 arguments at: 1 ifAbsent: [ ]).
				str = 'restore display (r)' ifTrue: [ mc2 receiver
						addLine;
						add: '**CLICK ME**' target: self selector: #removeHookExample1 ] ]
			named: #hookExample1
			inCategory: #menuAdditions
			fromClass: TheWorldMenu
			selector: #fillIn:from:.
! !

!Object class methodsFor: 'hooks-examples' stamp: 'nk 3/13/2001 21:39'!
hookExample2
	"This adds an item to the World Menu that then removes itself when clicked
	and then skips 'restore display' item in the menu by using terminateTo:"

	"Object hookExample2"
	"Hooks explore"

	"This block will be called with:
	cat		#menuAdditions
	tag		#hookExample1
	mc1		TheWorldMenu>>fillIn:from: context
	mc2		menu add* method contexts"

	self addHook: [ :cat :tag :mc1 :mc2 | | str |
				str _ (mc2 arguments at: 1 ifAbsent: [ ]).
				str = 'restore display (r)' ifTrue: [ mc2 receiver
						addLine;
						add: '**CLICK ME**' target: self selector: #removeHookExample1.
					thisContext terminateTo: mc2 sender
				 ]
			]
			named: #hookExample1
			inCategory: #menuAdditions
			fromClass: TheWorldMenu
			selector: #fillIn:from:.
! !

!Object class methodsFor: 'hooks-examples' stamp: 'nk 3/13/2001 20:53'!
removeHookExample1
	"This removes the item added by hookExample1 from the World Menu"

	self removeHookNamed: #hookExample1 inCategory: #menuAdditions fromClass: TheWorldMenu selector: #fillIn:from:.
! !

!Object class methodsFor: 'hooks' stamp: 'nk 3/13/2001 21:02'!
addHook: aBlock named: tag inCategory: cat fromClass: fromClass selector: fromSelector
	"Add the given (4-argument) block to the global list of Hooks to be applied on demand from methods called from the given class and selector. Give the block the given tag (which could be anything).
	Blocks are evaluated with four arguments:
		cat
		tag
		mc1		method context of the method that called mc2
		mc2		method context of the method that called callHooksInCategory: (often a menu method)
	"

	(((Hooks at: cat ifAbsentPut: [ IdentityDictionary new ])
		at: fromClass ifAbsentPut: [ IdentityDictionary new ])
			at: fromSelector ifAbsentPut: [ IdentityDictionary new ])
				at: tag put: aBlock fixTemps.
! !

!Object class methodsFor: 'hooks' stamp: 'nk 3/13/2001 20:48'!
hookNamed: tag inCategory: cat fromClass: fromClass selector: fromSelector
	"Answer the named tag or nil"

	^(((Hooks at: cat ifAbsent: [ ^nil ])
		at: fromClass ifAbsent: [ ^nil ])
			at: fromSelector ifAbsent: [ ^nil ])
				at: tag ifAbsent: [ ]
! !

!Object class methodsFor: 'hooks' stamp: 'nk 3/13/2001 21:15'!
initializeHooks
	"Build the Hooks dictionary as an environment global"
	"Object initializeHooks"

	self class environment atOrAbove: #Hooks ifAbsent: [
		self class environment at: #Hooks put: IdentityDictionary new ].! !

!Object class methodsFor: 'hooks' stamp: 'nk 3/13/2001 21:10'!
removeHookNamed: tag inCategory: cat fromClass: fromClass selector: fromSelector
	"Remove the named tag and return it, or return nil.
	Clean up afterwards"

	| classDict selectorDict tagDict oldHook |
	classDict _ Hooks at: cat ifAbsent: [ ^nil ].
	selectorDict _ classDict at: fromClass ifAbsent: [ ^nil ].
	tagDict _ selectorDict at: fromSelector ifAbsent: [ ^nil ].
	oldHook _ tagDict removeKey: tag ifAbsent: [ ].
	tagDict isEmpty ifTrue: [ selectorDict removeKey: fromSelector ].
	selectorDict isEmpty ifTrue: [ classDict removeKey: fromClass ].
	classDict isEmpty ifTrue: [ Hooks removeKey: cat ].
	^oldHook
! !


!Celeste class methodsFor: 'instance creation' stamp: 'rmp 3/20/2001 20:03'!
current
	"Answer the currently active Celeste (assuming that there's
	only one Celeste open at a given time) or open a new one."

	^Celeste allInstances detect: [:e | e dependents size ~= 0 ] ifNone: [self open].



"	^Celeste allInstances detect: [:e | e isActive] ifNone: [self open]. "! !


!PianoRollScoreMorph methodsFor: 'drawing' stamp: 'rmp 2/7/2001 12:57'!
addNotes
	"Recompute the set of morphs that should be visible at the current scroll position."

	| visibleMorphs rightEdge topEdge trackColor i done n nLeft nTop nRight rightEdgeTime |
	visibleMorphs _ OrderedCollection new: 500.
	rightEdge _ self right - borderWidth.
	rightEdgeTime _ self timeForX: rightEdge.
	topEdge _ self top + borderWidth + 1.

	"Add ambient morphs first (they will be front-most)"
	score eventMorphsWithTimeDo:
		[:m :t | m addMorphsTo: visibleMorphs pianoRoll: self eventTime: t
					betweenTime: leftEdgeTime and: rightEdgeTime].

	"Then add note morphs"
	score tracks withIndexDo:
		[:track :trackIndex |
		trackColor _ colorForTrack at: trackIndex.
		trackColor alpha: 0.5. 
		i _ indexInTrack at: trackIndex.
		done _ scorePlayer mutedForTrack: trackIndex.
		[done | (i > track size)] whileFalse: [
			n _ track at: i.
			(n isNoteEvent and: [n midiKey >= lowestNote]) ifTrue: [
				n time > rightEdgeTime
					ifTrue: [done _ true]
					ifFalse: [
						nLeft _ self xForTime: n time.
						nTop _ (self yForMidiKey: n midiKey) - 1.
						nTop > topEdge ifTrue: [
							nRight _ nLeft + (n duration * timeScale) truncated - 1.
							visibleMorphs add:
								((PianoRollNoteMorph
									newBounds: (nLeft@nTop corner: nRight@(nTop + 3))
									color: trackColor)
									trackIndex: trackIndex indexInTrack: i)]]].
			i _ i + 1].
			(selection notNil
				and: [trackIndex = selection first
				and: [i >= selection second and: [(indexInTrack at: trackIndex) <= selection third]]])
				ifTrue: [visibleMorphs do:
						[:vm | (vm isKindOf: PianoRollNoteMorph) ifTrue: [vm selectFrom: selection]]]].

	"Add the cursor morph in front of all notes; height and position are set later."
	cursor ifNil: [cursor _ Morph newBounds: (self topLeft extent: 1@1) color: Color red].
	visibleMorphs addFirst: cursor.

	self changed.
	self removeAllMorphs.
	self addAllMorphs: visibleMorphs.
! !


!Scamper methodsFor: 'menus' stamp: 'rmp 3/20/2001 20:40'!
displayDocument: mimeDocument
	"switch to viewing the given MIMEDocument"
	|  handled  urlString |
	handled _ false.


	"add it to the history"
"	recentDocuments removeAllSuchThat: [ :d | d url = mimeDocument url ]."
	currentUrlIndex > recentDocuments size
		ifTrue: [recentDocuments addLast: mimeDocument].
"		ifFalse: [recentDocuments removeAt: currentUrlIndex]."
	recentDocuments size > 20 ifTrue: [ recentDocuments removeFirst ].
		
	mimeDocument mainType = 'image' 
		ifTrue: [handled _ self displayImagePage: mimeDocument].

	mimeDocument contentType = 'text/html' 
		ifTrue: [handled _ self displayTextHtmlPage: mimeDocument].

	mimeDocument contentType = 'x-application/shockwave-flash'
		ifTrue:[handled _ self displayFlashPage: mimeDocument].

	(#('audio/midi' 'audio/x-midi') includes: mimeDocument contentType) 
		ifTrue: [handled _ self processMidiPage: mimeDocument].

	"Before we display plain text files we should look at the extension of the URL"
	(handled not and:[ mimeDocument mainType = 'text']) ifTrue:[
		urlString _ mimeDocument url toText.
		(urlString endsWithAnyOf: #('.gif' '.jpg' '.pcx')) 
			ifTrue:[handled _ self displayImagePage: mimeDocument].
		(handled not and:[urlString endsWithAnyOf: #('.mid' '.midi')])
			ifTrue:[handled _ self processMidiPage: mimeDocument].
		(handled not and:[urlString endsWith: '.swf'])
			ifTrue:[handled _ self displayFlashPage: mimeDocument].
	].

	(handled not and: [ mimeDocument mainType = 'text']) ifTrue: [
		self displayPlainTextPage: mimeDocument.
		handled _ true].


	handled ifFalse: [self processUnhandledPage: mimeDocument].

	self trigger: #scamperChangePage.
! !


!Scamper class methodsFor: 'initialization' stamp: 'rmp 3/20/2001 19:27'!
current
	"Answer the currently active Scamper (assuming that there's 
	only one Scamper open at a given time) or open a new one."

	^Scamper allInstances detect: [:e | e dependents size ~= 0 ] ifNone: [self open].
! !


!ScorePlayer methodsFor: 'accessing' stamp: 'rmp 3/13/2001 21:11'!
panForTrack: i

	| left right fullVol pan |
	left _ leftVols at: i.
	right _ rightVols at: i.
	left = right ifTrue: [^ 0.5].  "centered"
	fullVol _ left max: right.
	left < fullVol
		ifTrue: [pan _ 1.0 - (left asFloat / (2.0 * fullVol))]
		ifFalse: [pan _ (right asFloat / (2.0 * fullVol))].
	^ pan roundTo: 0.001

! !

!ScorePlayer methodsFor: 'volume' stamp: 'rmp 3/13/2001 20:58'!
panForTrack: trackIndex put: aNumber
	"Set the left-right pan for this track to a value in the range [0.0..1.0], where 0.0 means full-left."

	| fullVol pan left right |
	trackIndex > leftVols size ifTrue: [^ self].
	fullVol _ (leftVols at: trackIndex) max: (rightVols at: trackIndex).
	pan _ (aNumber asFloat min: 1.0) max: 0.0.
	pan <= 0.5
		ifTrue: [  "attenuate right channel"
			left _ fullVol.
			right _ 2.0 * pan * fullVol]
		ifFalse: [  "attenuate left channel"
			left _ 2.0 * (1.0 - pan) * fullVol.
			right _ fullVol].
	rightVols at: trackIndex put: right asInteger.
	leftVols at: trackIndex put: left asInteger.
! !


!ScorePlayerMorph methodsFor: 'initialization' stamp: 'rmp 3/13/2001 18:25'!
onScorePlayer: aScorePlayer title: scoreName
	| divider col r |
	scorePlayer _ aScorePlayer.
	scorePlayer ifNotNil:
		[scorePlayer  reset.
		instrumentSelector _ Array new: scorePlayer score tracks size].
	divider _ AlignmentMorph new
		extent: 10@1;
		borderWidth: 1;
		layoutInset: 0;
		borderColor: #raised;
		color: color;
		hResizing: #spaceFill;
		vResizing: #rigid.

	self removeAllMorphs.
	self addMorphBack: self makeControls.
	scorePlayer ifNil: [^ self].

	r _ self makeRow
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.
	r addMorphBack: self rateControl;
		addMorphBack: (Morph newBounds: (0@0 extent: 20@0) color: Color transparent);
		addMorphBack: self volumeControl.
	self addMorphBack: r.
	self addMorphBack: self scrollControl.

	col _ AlignmentMorph newColumn color: color; layoutInset: 0.
	self addMorphBack: col.
	1 to: scorePlayer trackCount do: [:trackIndex |
		col addMorphBack: divider fullCopy.
		col addMorphBack: (self trackControlsFor: trackIndex)].

	LastMIDIPort ifNotNil: [
		"use the most recently set MIDI port"
		scorePlayer openMIDIPort: LastMIDIPort].
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'rmp 3/13/2001 17:41'!
openMIDIFile
	"Open a MIDI score and re-init controls..."
	| fileName |
	fileName _ Utilities chooseFileWithSuffixFromList: #('.mid' '.midi')
					withCaption: 'Choose a MIDI file to open'.
	fileName ifNil: [^ self].
	self openMIDIFile: fileName.	
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'rmp 4/18/2001 14:39'!
openMIDIFile: fileName
	"Open a MIDI score and re-init controls..."
	| score f player |

	soundFileName _ fileName.
	(soundFileName notNil)
	ifTrue: [
		f _ FileStream readOnlyFileNamed: soundFileName.
		score _ (MIDIFileReader new readMIDIFrom: f binary) asScore.
		f close.
		player _ ScorePlayer onScore: score.
		soundFileName _ fileName.
		self onScorePlayer: player title: fileName	
	]
	ifFalse: [
		score _ nil.
		self onScorePlayer: (ScorePlayer new initialize) title: ' '.
	].! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'rmp 3/13/2001 19:11'!
refresh
	| divider col r |
	scorePlayer ifNotNil:
		[instrumentSelector _ Array new: scorePlayer score tracks size].
	divider _ AlignmentMorph new
		extent: 10@1;
		borderWidth: 1;
		layoutInset: 0;
		borderColor: #raised;
		color: color;
		hResizing: #spaceFill;
		vResizing: #rigid.

	self removeAllMorphs.
	self addMorphBack: self makeControls.
	scorePlayer ifNil: [^ self].

	r _ self makeRow
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.
	r addMorphBack: self rateControl;
		addMorphBack: (Morph newBounds: (0@0 extent: 20@0) color: Color transparent);
		addMorphBack: self volumeControl.
	self addMorphBack: r.
	self addMorphBack: self scrollControl.

	col _ AlignmentMorph newColumn color: color; layoutInset: 0.
	self addMorphBack: col.
	1 to: scorePlayer trackCount do: [:trackIndex |
		col addMorphBack: divider fullCopy.
		col addMorphBack: (self trackControlsFor: trackIndex)].

	LastMIDIPort ifNotNil: [
		"use the most recently set MIDI port"
		scorePlayer openMIDIPort: LastMIDIPort].
! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'rmp 3/13/2001 17:03'!
logOfRate

	^ LogOfRate.! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'rmp 3/14/2001 13:34'!
muteButtonForTrack: trackIndex
	"if the index is out of the range exit. "
	((trackIndex * 2) > (self submorphs at: 4) submorphs size )
	ifTrue: [ ^ nil ].
	
	^ ((((self submorphs at: 4) submorphs at: (2 * trackIndex)) submorphs at: 1) submorphs at: 7).! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'rmp 3/14/2001 12:53'!
overallVolumeMorph

	^ (((self submorphs at: 2) submorphs at: 3) submorphs at: 2)! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'rmp 3/14/2001 13:31'!
panMorphForTrack: trackIndex
	"if the index is out of the range exit. "
	((trackIndex * 2) > (self submorphs at: 4) submorphs size )
	ifTrue: [ ^ nil ].
	
	^ (((((self submorphs at: 4) submorphs at: (2 * trackIndex)) submorphs at: 3) submorphs at: 2) submorphs at: 2).! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'rmp 3/14/2001 12:56'!
rateSliderMorph

	^ (((self submorphs at: 2) submorphs at: 1) submorphs at: 2)! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'rmp 3/13/2001 16:53'!
soundFileName

	^ soundFileName.! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'rmp 3/14/2001 13:28'!
volumeMorphForTrack: trackIndex
	"if the index is out of the range exit. "
	((trackIndex * 2) > (self submorphs at: 4) submorphs size )
	ifTrue: [ ^ nil ].
	
	^ (((((self submorphs at: 4) submorphs at: (2 * trackIndex)) submorphs at: 3) submorphs at: 1) submorphs at: 2).! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'rmp 3/13/2001 20:10'!
trackNumAndMuteButtonFor: trackIndex

	| muteButton instSelector pianoRollColor r |
	muteButton _ SimpleSwitchMorph new
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		offColor: color;
		color: color;
		label: 'Mute';
		target: scorePlayer;
		actionSelector: #mutedForTrack:put:;
		arguments: (Array with: trackIndex).
	instSelector _ PopUpChoiceMorph new
		extent: 95@14;
		contentsClipped: 'oboe1';
		target: self;
		actionSelector: #atTrack:from:selectInstrument:;
		getItemsSelector: #instrumentChoicesForTrack:;
		getItemsArgs: (Array with: trackIndex).
	instSelector arguments:
		(Array with: trackIndex with: instSelector).
	instrumentSelector at: trackIndex put: instSelector.

	muteButton setSwitchState: (scorePlayer mutedForTrack: trackIndex). "If the tracks are muted the button should be depressed."

	"select track color using same color list as PianoRollScoreMorph"
	pianoRollColor _ (Color wheel: scorePlayer score tracks size) at: trackIndex.

	r _ self makeRow
		hResizing: #rigid;
		vResizing: #spaceFill;
		extent: 70@10.
	r addMorphBack:
		((StringMorph
			contents: trackIndex printString
			font: (TextStyle default fontOfSize: 24)) color: pianoRollColor).
	trackIndex < 10
		ifTrue: [r addMorphBack: (Morph new color: color; extent: 19@8)]  "spacer"
		ifFalse: [r addMorphBack: (Morph new color: color; extent: 8@8)].  "spacer"
	r addMorphBack:
		(StringMorph new
			extent: 140@14;
			contentsClipped: (scorePlayer infoForTrack: trackIndex)).
	r addMorphBack: (Morph new color: color; extent: 8@8).  "spacer"
	r addMorphBack: instSelector.
	r addMorphBack: (AlignmentMorph newRow color: color).  "spacer"
	r addMorphBack: muteButton.
	^ r
! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'rmp 3/13/2001 18:48'!
instrumentSelector

	^ instrumentSelector.
! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'rmp 2/21/2001 11:16'!
invokeMenu
	"Invoke a menu of additonal functions for this WaveEditor."
	| aMenu |
	aMenu _ MenuMorph new defaultTarget: self.
	aMenu add: 'reload instruments' target: AbstractSound selector: #updateScorePlayers.
	aMenu add: 'open a MIDI file' action: #openMIDIFile.
	scorePlayer midiPort
		ifNil: [
			aMenu add: 'play via MIDI' action: #openMIDIPort]
		ifNotNil: [
			aMenu add: 'play via built in synth' action: #closeMIDIPort.
			aMenu add: 'new MIDI controller' action: #makeMIDIController:].
	aMenu addLine.
	aMenu add: 'make a pause marker' action: #makeAPauseEvent:.

	aMenu popUpInWorld: self world.! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'rmp 3/13/2001 17:03'!
setLogRate: logOfRate

	LogOfRate _ logOfRate.
	scorePlayer rate: (3.5 raisedTo: logOfRate).
! !


!ScorePlayerState methodsFor: 'as yet unclassified' stamp: 'rmp 4/17/2001 18:27'!
getState: scorePlayerMorph
	"Assign self the current state of scorePlayerMorph."

	| size SP |

	SP _ scorePlayerMorph scorePlayer.
	FileName _ scorePlayerMorph soundFileName.
	LogOfRate _ scorePlayerMorph logOfRate.
	( LogOfRate isNil )
		ifTrue: [ LogOfRate _ 0.0 ] .
	Volume _ SP overallVolume.


	size _ SP trackCount.	
	InstrumentArray _ Array new: size.
	1 to: size do: [:i |
		InstrumentArray at: i put: ((scorePlayerMorph instrumentSelector at: i) contents).
	].
	
	VolumeArray _ Array new: size.
	1 to: size do: [:i |
		VolumeArray at: i put: (SP volumeForTrack: i).
	].

	BalanceArray _ Array new: size.
	1 to: size do: [:i |
		BalanceArray at: i put: (SP panForTrack: i).
	].

	MuteArray _ Array new: size.
	1 to: size do: [:i | 
		MuteArray at: i put: (SP mutedForTrack: i).
	].

	
	! !

!ScorePlayerState methodsFor: 'as yet unclassified' stamp: 'rmp 4/17/2001 16:51'!
setStateOf: scorePlayerMorph
	"comment stating purpose of message"

	| size SP |

	SP _ scorePlayerMorph scorePlayer.
	( FileName = (scorePlayerMorph soundFileName) )
	ifFalse: [
		scorePlayerMorph scorePlayer pause.
		scorePlayerMorph openMIDIFile: FileName.
		SP _ scorePlayerMorph scorePlayer.
	].
	(SP isPlaying)
	ifFalse: [
		scorePlayerMorph scorePlayer resumePlaying.
	].

	"set the rate"
	scorePlayerMorph setLogRate: LogOfRate.
	scorePlayerMorph rateSliderMorph adjustToValue: LogOfRate.

	"set the overall volume"
	SP overallVolume: Volume.
	scorePlayerMorph overallVolumeMorph adjustToValue: Volume.

	size _ SP score tracks size.	
	
	1 to: size do: [:i |
		SP volumeForTrack: i put: (VolumeArray at: i).
		(scorePlayerMorph volumeMorphForTrack: i) adjustToValue: (VolumeArray at: i).
	].

	1 to: size do: [:i |
		SP panForTrack: i put: (BalanceArray at: i).
		(scorePlayerMorph panMorphForTrack: i) adjustToValue: (BalanceArray at: i).
	].

	1 to: size do: [:i | 
		 SP mutedForTrack: i put: (MuteArray at: i).
		(scorePlayerMorph muteButtonForTrack: i) setSwitchState: (MuteArray at: i).
	].

	1 to: size do: [:i |
		scorePlayerMorph atTrack: i from: nil selectInstrument: (InstrumentArray at: i).
	].

! !

ScorePlayerState class removeSelector: #setStateOf:with:!
ScorePlayerState removeSelector: #setStateInstruments:!
ScorePlayerState removeSelector: #setStateMuted:!
ScorePlayerMorph removeSelector: #refreshPlayer:!
AlignmentMorph subclass: #ScorePlayerMorph
	instanceVariableNames: 'LogOfRate soundFileName scorePlayer trackInstNames instrumentSelector scrollSlider '
	classVariableNames: 'LastMIDIPort '
	poolDictionaries: ''
	category: 'Sound-Scores'!
MusicMonitor class removeSelector: #scorePlayerMorph!
MusicMonitor removeSelector: #checkCeleste!
MusicMonitor removeSelector: #checkMessage!
MusicMonitor removeSelector: #checkScamper!
MusicMonitor removeSelector: #compiler!
MusicMonitor removeSelector: #currentconditional!
MusicMonitor removeSelector: #scorePlayerState:condition:!
MusicMonitor removeSelector: #scorePlayerStop!
MMFilter removeSelector: #filename!
MMFilter removeSelector: #filename:!
MMFilter removeSelector: #instrumentArray!
MMFilter removeSelector: #instrumentArray:!
MMFilter removeSelector: #triggerString!
MMFilter removeSelector: #triggerString:!
Object removeSelector: #greeting!
"Postscript:
Initialize the Hooks global"
Object initializeHooks!

